extends KinematicBody2D

signal hit

export var NUMBER_OF_SENSORS = 8
var max_range_sensor = 1000; # sensor range in pixels (31m)
var range_per_step_sensor = 10; # accuracy in pixels (0.31m)
var brake_str = 1 # reduces braking power

var reverse_acceleration = 0.5 # reverse acceleration in correlation to forward acceleration
var max_speed_reverse = 250 # in pixels (28 km/h)

#var delta2 = 0.017
var current_speed=0 # absolute speed of car
onready var start_time = OS.get_ticks_msec()
var lauf = OS.get_ticks_msec()
var turn = 0
var max_angle = 22 # max angle so the turn-radius is 11m (VW Golf R 2015 for reference)
var direction # determines if car is driving forwards or backwards



var wheel_base = 70 # distance between wheels in pixels
#var steering_angle = 0.1 # 
var engine_power = 1 # can be lowered if you want the car to accelerate and brake slower
var braking = -360*2 # braking power


export var score = 250

var acceleration = Vector2.ZERO
export var velocity = Vector2.ZERO
export var steer_direction = 0



var scoreTimer = 1 
export var scoreTimerFreeze = false #toogled when finish Line is reached to stop lowering the score
	
export var distance_sensors= [] #array with distance of distance Sensors
export var linear_velocityN = Vector2.ZERO


func reset():
	#called from world on reset
	#resets variables to start Values
	acceleration = Vector2.ZERO
	velocity = Vector2.ZERO
	steer_direction = 0
	score = 250
	scoreTimerFreeze = false 
	turn = 0
	
func _ready():
	#turn of process so it time can be controlled from time method in world
	set_physics_process(false)
	_network_process(0.001)
	
	#prepare distance sensors array
	for _i in range(NUMBER_OF_SENSORS):
		distance_sensors.append(Vector2.ZERO)

func _network_process(delta):
	# calls all actions of a step

	_score_Timer(delta)
	acceleration = Vector2.ZERO
	calc_steering(delta)
	calc_acceleration(delta)
	calc_orientation(delta)
	velocity += acceleration * delta
	if self.move_and_collide(delta*velocity):
		velocity=Vector2.ZERO
	linear_velocityN = velocity
	calc_distance_Sensor()



func calc_distance_Sensor():
	# get global position of car
	var t = self.global_transform
	
	#calculate rotation of each sensors
	for j in range(len(distance_sensors)):
		var rot = 2*PI/len(distance_sensors)*j
		
		#test for collision on the given direction of the sensor
		var i = 0
		while (!test_move(self.global_transform,i*t.rotated(rot).basis_xform(Vector2(1,0)).normalized(),false)&&i<max_range_sensor): #t.rotated(rot).basis_xform(Vector2(1,1))
			i = i+range_per_step_sensor
		
		#save max distance without collision
		distance_sensors[j] = i*t.rotated(rot).basis_xform(Vector2(1,0)).normalized()
		
	

func calc_acceleration(delta):	
	# calculates absolute speed from velocity vector
	current_speed = sqrt(pow(velocity[0],2)+pow(velocity[1],2))/32
	
	# calculates acceleration in correlation to current speed (VW Golf R 2015 for reference)
	if Input.is_action_pressed("accelerate"):
		acceleration= transform.x * (0.0001*pow(current_speed,3)-0.0093*pow(current_speed,2)+0.0802*current_speed+7.0157)
		acceleration=acceleration*32*engine_power
		
	# brakes at brake_str (currently 0.5) of emergency braking
	if Input.is_action_pressed("brake"):
		if  direction > 0:
#			if brake_str < 1:
#				brake_str += 0.01
			#brake_str = 0.2
			acceleration = transform.x * (braking * engine_power)*brake_str
			if current_speed*32 - acceleration.length() * delta < 0:
				velocity = Vector2.ZERO
				acceleration = Vector2.ZERO
				linear_velocityN = velocity
		else:
#			if brake_str < 1:
#				brake_str += 0.01
			#brake_str = 0.2
			acceleration = -transform.x * (braking * engine_power) * brake_str
			if -current_speed*32 + acceleration.length() * delta > 0:
				velocity = Vector2.ZERO
				acceleration = Vector2.ZERO
				linear_velocityN = velocity
#	else:
#		brake_str = 0

	# can only reverse while below 5km/h
	# reverses at reverse_acceleration (currently 0.5) of forward acceleration 
	if Input.is_action_pressed("reverse"):
		if current_speed*direction < 5:
			if current_speed < max_speed_reverse:
				acceleration= transform.x * (0.0001*pow(current_speed,3)-0.0093*pow(current_speed,2)+0.0802*current_speed+7.0157)
				acceleration= -acceleration*32*engine_power*reverse_acceleration
	if lauf<OS.get_ticks_msec():
		lauf += 1000
		#print("Elapsed time: ", OS.get_ticks_msec() - start_time)
		#print((current_speed+1)*3.6)
		
func calc_steering(delta):
	# determines angle of the virtual wheel
	var lenken = delta*59
	if Input.is_action_pressed("steer_right"):
		turn += 0.2*lenken
	elif Input.is_action_pressed("steer_left"):
		turn -= 0.2*lenken
#	else:
#		 #überspringt Lenkrückstellung einkommentieren um wieder zu aktivieren
#		if turn > 0:
#			turn -= 2*lenken
#		if turn < 0:
#			turn += 2*lenken
	if turn < -max_angle:
		turn = -max_angle
	if turn > max_angle:
		turn = max_angle
	steer_direction = turn #* steering_angle
	steer_direction = deg2rad(steer_direction)

func calc_orientation(delta):
	# calculates the new orientation of the car after turning
	var rear_wheel = position - transform.x * wheel_base/2.0
	var front_wheel = position + transform.x * wheel_base/2.0
	rear_wheel += velocity * delta
	front_wheel += velocity.rotated(steer_direction) * delta
	var new_heading = (front_wheel - rear_wheel).normalized()
	
	direction = new_heading.dot(velocity.normalized())
	if direction > 0:
		velocity = new_heading * velocity.length()
	if direction < 0:
		velocity = -new_heading * min(velocity.length(), max_speed_reverse)
	linear_velocityN = velocity
	rotation = new_heading.angle()





func _score_Timer(delta):
	#used to lower the score by 1 every second. Since update lowers the score by delta every step
	if !scoreTimerFreeze:
		scoreTimer -= delta
		score -= delta

	if scoreTimer <= 0:
		scoreTimer = 1
		

	




func _on_Area2D_body_entered(body):
	emit_signal("hit")
	
	#check if collision cooldwon is inactive
#	if collisionTimer <= 0:
		
	#if hit by mob lower the score by 48
	if body is RigidBody2D:
		#print("Kollision: -50 Punkte")
		score -= 48
		
	#if hit by TileMap (Wall) lower the score by 52
	if body is TileMap:
		score -= 52
	
	#collisionTimer = 1 #set collision cooldwon to 1


