extends Node2D

export(PackedScene) var Mob
export(PackedScene) var Mob_LKW

signal won
# The port we will listen to.
const PORT = 9080
# Our WebSocketServer instance.
var _server = WebSocketServer.new()
var client_id
#var car_data = new PoolByteArray
#var json = JSON.new()
export var deltaN = 0.001



var min_speed = 700
var max_speed = 1100
var lines = []
var oldScore = 250

var client
var server
var wrapped_client
var connected = false
var message_center
var should_connect = false

var activeMap = 1

var lane1Timer = 1
var lane2Timer = 1
var done = false

export var actions = []

var human_player = false
export var learn_with_images = false
export var show_game = true

func _ready():
	
	#Toggle game visabillity
	if human_player:
		show_game = true
	if learn_with_images:
		show_game = true
		$CanvasLayer/ControlPanel.hide()
	if not show_game:
		VisualServer.render_loop_enabled = false # disable rendering to create a massive boost
	#Remove Control pannel to Optimize Image for Maschine learning

		
	# Connect base signal to get notified of new client connections
	_server.connect("client_connected", self, "_connected")

	
	# This signal is emitted when not using the Multiplayer API every time a
	# full packet is received.
	_server.connect("data_received", self, "_on_data")
	
	# Start listening on the given port.
	var err = _server.listen(PORT)
	if err != OK:
		print("Unable to start server")
		set_process(false)
	
	
	# Chose Map and move the car to start position an rotation
	prepareMap()
	

func prepareMap():
	get_tree().call_group("mobs","queue_free")
	
	#Random Choice of Map
	if Input.is_action_pressed("map1"):
		activeMap = 1
	elif Input.is_action_pressed("map2"):
		activeMap=2
	elif Input.is_action_pressed("map3"):
		activeMap=3
	else:
		activeMap = randi()%3+1
	
	#Coment in with Map of your Choice to supress Random 
	#activeMap = 3
	
	#Place Spawn points and Finish Line at chosen Map
	if activeMap == 1: 
		$Lane1Spawn.global_position = $map1Lane1Spawn.global_position
		$Lane2Spawn.global_position = $map1Lane2Spawn.global_position
		$PlayerSpawn.global_position = $map1CarSpawn.global_position
		$PlayerSpawn.global_rotation = $map1CarSpawn.global_rotation
		$Finish_line/CollisionShape2D.global_position = $map1FinishLine.global_position
		$StartTimer.start()
		
	elif activeMap == 2:
		$Lane1Spawn.global_position = $map2Lane1Spawn.global_position
		$Lane2Spawn.global_position = $map2Lane2Spawn.global_position
		$PlayerSpawn.global_position = $map2CarSpawn.global_position
		$PlayerSpawn.global_rotation = $map2CarSpawn.global_rotation
		$Finish_line/CollisionShape2D.global_position = $map2FinishLine.global_position
		$StartTimer.start()
	
	elif activeMap == 3:
		$Lane1Spawn.global_position = $map3Lane1Spawn.global_position
		$Lane2Spawn.global_position = $map3Lane2Spawn.global_position
		$PlayerSpawn.global_position = $map3CarSpawn.global_position
		$PlayerSpawn.global_rotation = $map3CarSpawn.global_rotation
		$Finish_line/CollisionShape2D.global_position = $map3FinishLine.global_position
		$StartTimer.start()
	
	#Move Car to spawn position and rotation
	$Car.global_position = $PlayerSpawn.global_position
	$Car.global_rotation = $PlayerSpawn.global_rotation

func _draw():
	#Draw distance sensors Lines
	if not learn_with_images:
		if $Car.distance_sensors:
			for i in range($Car.NUMBER_OF_SENSORS):
				draw_line($Car.global_position,$Car.global_position+$Car.distance_sensors[i],Color(255, 0, 0), 1)

func _process(delta):
	
	#Run time if a human is playing
	if human_player:
		time(delta/2)
		
	# Call server to check for new Connections and Data
	_server.poll()

	#Update the Canvas
	update()
	
	
func set_camera_limits():
	# Sets the boundaries of the field of view
	var map_limits = $TileMap.get_used_rect()
	var map_cellsize = $TileMap.cell_size
	$CarTest/Camera2D.limit_left = map_limits.position.x * map_cellsize.x
	$CarTest/Camera2D.limit_top = map_limits.position.y * map_cellsize.y
	$CarTest/Camera2D.limit_right = map_limits.end.x * map_cellsize.x
	$CarTest/Camera2D.limit_bottom = map_limits.end.y * map_cellsize.y

func _on_Lane1Timer_timeout():
	
	#Create new Instance of Mob
	var mob = Mob.instance()
	#Add Mob instance to World
	add_child(mob)
	#Move Mob to its spawn position an rotation
	mob.position = $Lane1Spawn.position
	mob.rotation = PI
	
	#Set mob speed and save it as standard speed
	mob.speed = rand_range(min_speed,max_speed)
	mob.standard_speed = mob.speed
	
	#Set spawn Timer for next Mob to a random Time between 1 and 2 Seconds
	lane1Timer = rand_range(1.0,2.0)

func _on_Lane2Timer_timeout():
	#Create new Instance of Mob or LKW_Mob randomly chosen and set the speed of the Instance
	var rand = rand_range(0,2)
	var mob
	if(rand>1.5):
		mob = Mob.instance()
		mob.speed = rand_range(min_speed,max_speed)
	else:
		mob = Mob_LKW.instance()
		mob.speed = rand_range(min_speed-200,max_speed-400)
	#Save speed as standard speed
	mob.standard_speed = mob.speed
	
	#Add Mob instance to World
	add_child(mob)
	#Move Mob to its spawn position an rotation
	mob.position = $Lane2Spawn.position
	mob.rotation = PI
	
	#Set spawn Timer for next Mob to a random Time between 1 and 2 Seconds
	lane2Timer = rand_range(1.0,2.0)


func _on_Finish_line_body_entered(body):
	#If Body is Car freeze score Timer and raise score
	if body is KinematicBody2D:
		emit_signal("won")
		body.scoreTimerFreeze = true
		body.score += 200
		
	#If Body is Mob remove from World
	if body is RigidBody2D:
		body.queue_free()


func _on_World_won():
	#Function called when Player enters Finish_Line Area
	done = true
	
	
func _on_StartTimer_timeout():
	
	#move Lane 1 and Lane 2 spawns down by a random distance
	$Lane1Spawn.global_position = $Lane1Spawn.global_position + Vector2(0,rand_range(1280,1920))
	$Lane2Spawn.global_position = $Lane2Spawn.global_position + Vector2(0,rand_range(1280,1920))
	
	#spawn Mobs on new position
	_on_Lane1Timer_timeout()
	_on_Lane2Timer_timeout()
	
	#if Lane 1 or Lane 2 Spawn ar below 4000 move them to final position and stop start timer
	if($Lane2Spawn.global_position.y>4000):
		$Lane1Spawn.global_position.y = 5000
		$Lane2Spawn.global_position.y = 5000
		$StartTimer.stop()

func _connected(id, proto):
	# This is called when a new peer connects, "id" will be the assigned peer id,
	# "proto" will be the selected WebSocket sub-protocol (which is optional)
	print("Client %d connected with protokol: %s" % [id, proto])
	client_id = id
	

func _close_request(id, code, reason):
	id.close()
	# This is called when a client notifies that it wishes to close the connection,
	# providing a reason string and close code.
	print("Client %d disconnecting with code: %d, reason: %s" % [id, code, reason])
	pass
#
func _disconnected(id, was_clean = false):
#	# This is called when a client disconnects, "id" will be the one of the
#	# disconnecting client, "was_clean" will tell you if the disconnection
#	# was correctly notified by the remote peer before closing the socket.
	print("Client %d disconnected, clean: %s" % [id, str(was_clean)])



func _on_data(id):
	
	#set all input events to false
	for i in InputMap.get_actions():
		var a = InputEventAction.new()
		a.action = i
		a.pressed = false
		Input.parse_input_event(a)
	
	#get the incomming packet
	var pkt = _server.get_peer(id).get_packet()
	
	#parse commands form packet to array
	var commands = (JSON.parse(pkt.get_string_from_utf8())).get_result()
	
	
	if(commands):
		#in commands arry on the first position is a array of strings 
		#with input commands these correspond to the input events
		for i in commands[0]:
			var a = InputEventAction.new()
			a.action = i
			a.pressed = true
			Input.parse_input_event(a)
		# on the second position is the time to be run
		deltaN=commands[1]
	
	# run the programm by the given time
	time(deltaN)
	
	# get the Sensors values and other Car Data as utf8_JSON package
	#and send the package back to the client
	var car_data = yield(get_car_data_as_utf8_JSON(),"completed")
	_server.get_peer(id).put_packet(car_data)
	
	#Save the score to calculate the Score Delta in the next round
	oldScore = $Car.score
	
func time(delta):
	#save delta as DeltaN in case time() is not called from _on_Data
	
	deltaN = delta
	
	#create actions array:
	if show_game:
		actions=[]
		for i in InputMap.get_actions():
			if Input.is_action_pressed(i):
				actions.append(i)
	
	# enable rendering 
	if Input.is_action_pressed("screen_on"):
		VisualServer.render_loop_enabled = true 
		
	# disable rendering to create a massive boost
	if Input.is_action_pressed("screen_off"):
		VisualServer.render_loop_enabled = false 	
	
	#reset variables and move car to start position, notice that Mobs do not get reset
	if Input.is_action_pressed("reset"):
		prepareMap()
		$Car.reset()
		done = false
	
	#run Timers by given Time
	runTimer(delta)
	
	#run Car by given Time
	$Car._network_process(delta)
	
	#run Canvas by given Time
	$CanvasLayer/ControlPanel._network_process(delta)
	
	#run Mobs by given Time. Since ther are serveral Mobs they have to be called as a Group
	get_tree().call_group("mobs","_call_process")
	

func runTimer(delta):
	#subtract time from timer
	lane1Timer=lane1Timer-delta
	lane2Timer=lane2Timer-delta
	
	#check if timers have timed out
	if lane1Timer<=0:
		_on_Lane1Timer_timeout()
	if lane2Timer<=0:
		_on_Lane2Timer_timeout()

func _exit_tree():
	#stop server when game stops
	_server.stop()




func _on_NetworkTimer_timeout():
	#if played manually data can be send to a client through this function
	var car_data = yield(get_car_data_as_utf8_JSON(),"completed")
	if(client_id):
		_server.get_peer(client_id).put_packet(car_data)
	oldScore = $Car.score
	
func get_car_data_as_utf8_JSON():
	#Calculate distance of distance Sensors in meters
	var car_distance_in_m = []
	for i in $Car.distance_sensors:
		car_distance_in_m.append(i.length()/32)
	
	#take a screenshot for learning with images
	var screenshot = yield(get_screenshot(), "completed")
	
	#coment next line in to see the data send.
	#print([$Car.velocity.length()/32*3.6*$Car.d,rad2deg($Car.steer_direction),car_distance_in_m,[$Car.score,oldScore-$Car.score],done])
	

	#create array with Car Data
	var data = [$Car.velocity.length()/32*3.6*$Car.direction,rad2deg($Car.steer_direction),car_distance_in_m,[$Car.score,oldScore-$Car.score],done,screenshot,actions]
	
	#parse array to UTF8 JSON and return it
	return (JSON.print(data)).to_utf8()



func get_screenshot():
	
	if learn_with_images:
		get_viewport().set_clear_mode(Viewport.CLEAR_MODE_ONLY_NEXT_FRAME)
		# Wait until the frame has finished before getting the texture.
		yield(VisualServer, "frame_post_draw")

		var thumbnail = get_viewport().get_texture().get_data()
		thumbnail.flip_y()
		# Create a texture for it.

		#resize Image to smaler resolution
		thumbnail.resize(60, 80 )
		
		#coment in next Line and out the line after if you want the Image to be returned in color
		
		#thumbnail.convert(Image.FORMAT_RGB8 ) # Farbe
		thumbnail.convert(Image.FORMAT_L8 ) # S/W
		
		#thumbnail.save_png('test.png') # Save Image as file - to debug
		
		#get Data of the Image to be returned
		var array : PoolByteArray = thumbnail.get_data()
		var sg_width = thumbnail.get_width()
		var sg_height = thumbnail.get_height()
		var sg_format = thumbnail.get_format()
		var sg_u_size = array.size()
		
		#convert image array to base64 to be able to send it through JSON
		var sg_saved_img = Marshalls.raw_to_base64(array)
		
		# save Metadata in JSON format
		var metadata = {"thumbnail" : {"image" : sg_saved_img, "size" : sg_u_size, "width" : sg_width, "height" : sg_height, "format" : sg_format}}
		
		return metadata
	else:
		#if learn_with:images is turned of return Metadate with Image disabled
		var metadata = {"thumbnail" : {"image" : "disabled", "size" : 0, "width" : 0, "height" : 0, "format" : "diabled"}}
		yield(get_tree().create_timer(0),"timeout") # Hack
		return metadata

